import argparse
import os
import re

from metadocker.docker import DockerBuilder
from metadocker.env.base import Env, VarStr, VarInt, VarPath, VarDict
from metadocker.register import register_env



@register_env
class PythonServerEnv(Env):
    _name_ = "PythonServer"
    _vars_ = [

        VarStr("PROJECT", required=True, help="Le nom du projet", is_meta=True),
        VarStr("PROJECT_VERSION", required=True, help="La version du projet", is_meta=True, null=True),

        VarDict("ENV", {}, required=True, help="La liste des variable d'environnemnt à utiliser"),
        VarDict("PORTS", {8101:800}, required=True, help="Le mapping des ports. En clé les ports de l'hote et en valeur les port du container"),
        VarDict("VOLUMES", {"./data": "/data"}, required=True, help="Le mapping des ports. En clé les ports de l'hote et en valeur les port du container", is_meta=True, null=True),
        VarDict("COPY", {}, required=True,  help="Dictionaire des fichier a copier: cle: fs hote valeur fs invite"),
        VarPath("APP_DIR", "/data", required=True, help="Le dossier ou est lancé le projet"),

        VarStr("PYTHON_VERSION", "3.11", help="La version de python à utiliser"),
        VarStr("GIT_REPO", help="L'adresse du dépot git"),
        VarStr("GIT_VERSION", help="La version (branche ou tag) à utiliser. Utiliser None pour rester sur le master", null=True),
        VarStr("USER", os.environ["USER"], help="Le nom d'utilisateur à utiliser. L'utilisateur sera crée. Utiliser None pour rester en root"),#
        VarInt("UID", os.getuid(), help="L'UID de l'utilisateur, uniquement si USER est défini"), #
        VarPath("BUILD_DIR", "/build", help="Le dossier au sein du container ou est buildé le projet"),

        VarStr("APT_PACKAGES", help="La liste des packages apt à installer (séparé par des espaces)"),

        VarStr("RUN_BEFORE_SETUP", help="Une commande a utiliser juste avant de lancer la commande d'installation du module python"),
        VarStr("RUN_AFTER_SETUP", help="Une commande a utiliser juste après avoir lancé la commande d'installation du module python"),
        VarStr("INSTALL_CMD", "pip install .", help="La commande à utiliser pour installer le package python"),

        VarStr("RUN_CMD", help="La commande à installer pour lancer l'application. Par défaut: python -m {PROJECT}"),
   ]


    def init(self, docker : DockerBuilder):
        env = self.get_env()
        meta = self.get_meta()
        docker.meta(meta)
        docker.init(
            env.PROJECT,
            env.PROJECT_VERSION
        )
        return env, meta

    def _get_packages(self, env):
        data = env.APT_PACKAGES
        packages = []
        if env.GIT_REPO:
            packages.append("git")

        if isinstance(data, ((tuple, list, set))):
            packages.extend(data)
        else:
            packages.extend(re.sub("\s\+", " ", data).split(" "))
        return packages


    def _get_envs(self, env):
        envs=[f'APP_DIR="{env.APP_DIR}"']
        if env.ENV:
            for k, v in env.ENV.items():
                envs.append(f'{k}=\"{v}\"')
        for k, v in env.items():
            envs.append(f'CFG_{k}=\"{v}\"')
        return "\\\n    ".join(envs)

    def generate_docker_file(self, docker : DockerBuilder):

        debug = self.debug
        env, meta = self.init(docker)


        docker.from_(f"python:{env.PYTHON_VERSION}-slim")

        docker.env(self._get_envs(env))

        if env.COPY:
            for k, v in env.COPY.items():
                docker.copy(k, v)

        if env.PORTS:
            docker.expose(*env.PORTS.values())


        if env.USER:
            docker.run(f'adduser --disabled-password --gecos "" --home "/home/{env.USER}"  '
                           f'--shell "/sbin/nologin"  --uid "{env.UID}"  {env.USER}')

        packages = self._get_packages(env)
        if packages:
            docker.run(f'apt update -y && apt install -y', *packages)

        docker.run("mkdir -p", env.BUILD_DIR, env.APP_DIR)

        if env.USER:
            docker.run("mkdir -p", f"/home/{env.USER}", " ".join(env.VOLUMES.values()))
            docker.run(f"chown -R {env.USER}:{env.USER}", env.BUILD_DIR, f"/home/{env.USER}",
                       " ".join(env.VOLUMES.values()), env.APP_DIR)

            if debug:
                docker.run(f"echo 'root:root' | chpasswd")
                docker.run(f"echo '{env.USER}:root' | chpasswd")

            docker.user(env.USER)
            docker.env("USER", env.USER)
            docker.env("PATH", f"/home/{env.USER}/.local/bin:$PATH")


        if env.VOLUMES:
            docker.volume(" ".join(env.VOLUMES.values()))

        docker.workdir(env.BUILD_DIR)

        # docker.run("mkdir -p $HOME/.pip", "&&",
        #            r"echo '[global]' > $HOME/.pip/pip.conf &&",
        #            r"echo 'index-url = http://download.zope.org/simple' >> $HOME/.pip/pip.conf &&",
        #            r"echo 'trusted-host = download.zope.org' >> $HOME/.pip/pip.conf &&",
        #            r"echo 'extra-index-url= https://data.gautrais.eu/fanch/stable' >> $HOME/.pip/pip.conf")


        if env.GIT_REPO:
            if env.GIT_VERSION:
                docker.run(f"git clone {env.GIT_REPO} . && git checkout {env.GIT_VERSION}")
            else:
                docker.run(f"git clone {env.GIT_REPO} . ")


            if env.RUN_BEFORE_SETUP:
                docker.run(env.RUN_BEFORE_SETUP)

            docker.run("pip install .")

            if env.RUN_AFTER_SETUP:
                docker.run(env.RUN_AFTER_SETUP)
        else:
            if env.RUN_BEFORE_SETUP:
                docker.run(env.RUN_BEFORE_SETUP)

            if env.PROJECT_VERSION:
                docker.run(f"pip install {env.PROJECT}=={env.PROJECT_VERSION}")
            else:
                docker.run(f"pip install {env.PROJECT}")

            if env.RUN_AFTER_SETUP:
                docker.run(env.RUN_AFTER_SETUP)


        docker.workdir(env.APP_DIR)

        if env.RUN_CMD:
            docker.cmd(env.RUN_CMD)
        else:
            docker.run(f"python -m {env.PROJECT} migrate")
            docker.cmd(f"python -m {env.PROJECT} run")


